#SUPPLEMENTARY INFORMATION S2
#R CODE FOR THE ANALYSES PERFORMED IN HERSKIND & RIEDE 2024 (JAS)

#Part 1: Generating skipgrams and calculating PMI
#Part 2: Making culture complex-specific subsets and generating PMI summary stats
#Part 3: Making frequency- and PMI-sorted stacked bar plots
#Part 4: Making a heatmap of bigrams by their associated PMI
#Part 5: Exporting the data tables

#Notes:
#In part 1, the code has to be re-run for each level of n-gram analysis
#(2, 3, and 4). Certain lines are specifically for when making the 3- and 4-gram
#data table. For part 2, part 1 (the skipgram generation) is performed once more
#for each of the culture complex-specific subsets.

library(quanteda)
library(readxl)
library(ggplot2)
library(dplyr)
library(data.table)
library(data.table)
library(extrafont)
font_import(pattern = "GIL", prompt = FALSE)

################################################################################
#Part 1: Generating skipgrams and calculating PMI###############################

#Loading the Excel file #CHANGE FILE NAME TO WHATEVER THE SI-DATA IS CALLED#
data <- read_excel("Herskind&Riede_S1.xlsx")

#Converting the text data to a corpus 
#Change 'data' to 'maglemosedata', 'kongemosedata', & 'ertebølledata' for the
#culture complex-specific analyses#
corpus <- corpus(data$Motifs_as_text)
#the 'NA is replace by empty string' message is due to one decorated object not 
#fitting with the motif classification scheme.

#Setting the 'n' and 'k' values
#Change n value to 3 and 4 for tri- and quadrigram analyses
n <- 4
k <- 0:13

# tokenizing the corpus
toks <- tokens(corpus)

# creating the k-skip-n-grams
ksngrams <- tokens_ngrams(toks, n = n, skip = k, concatenator = " ")

head(ksngrams)

# creating a document-feature matrix
dfm <- dfm(ksngrams)

# calculating the frequency of each k-skip-n-gram:
freqs <- colSums(dfm)

head(freqs)

# Printing the total number of k-skip-n-grams
cat("Total number of ksngrams:", length(freqs), "\n")

# Counting the frequency of each frequency value (this is for table 1)
freq_count <- table(freqs)

freq_count

#Calculating individual token frequencies
dfm_toks <- dfm(toks)
token_freqs <- colSums(dfm_toks)

token_freqs

#Creating a data frame with k-skip-n-grams and their observed frequencies
ksngram_freq <- data.frame(ksngrams = names(freqs), observed_freq = freqs)

head(ksngram_freq)

ksngram_freq$token1 <- sapply(strsplit(ksngram_freq$ksngrams, " "), `[`, 1)
ksngram_freq$token2 <- sapply(strsplit(ksngram_freq$ksngrams, " "), `[`, 2)
##NEXT LINE IF RUNNING TRIGRAMS##
ksngram_freq$token3 <- sapply(strsplit(ksngram_freq$ksngrams, " "), `[`, 3)
##NEXT LINE IF RUNNING QUADRIGRAMS##
ksngram_freq$token4 <- sapply(strsplit(ksngram_freq$ksngrams, " "), `[`, 4)

#Creating a vector of all tokens
all_tokens <- names(token_freqs)

#Adding columns for token frequencies
ksngram_freq$token1_freq <- token_freqs[match(ksngram_freq$token1, all_tokens)]
ksngram_freq$token2_freq <- token_freqs[match(ksngram_freq$token2, all_tokens)]
##NEXT LINE IF WE'RE RUNNING TRIGRAMS##
ksngram_freq$token3_freq <- token_freqs[match(ksngram_freq$token3, all_tokens)]
##NEXT LINE IF WE'RE RUNNING QUADRIGRAMS##
ksngram_freq$token4_freq <- token_freqs[match(ksngram_freq$token4, all_tokens)]

head(ksngram_freq)

#Set to 482 rather than the actual number of rows/objects (483) because of the one
#empty string ("painted bands and fields")
t <- 482

# Calculating the expected frequency for each ksngram
ksngram_freq$expected_freq <- ((ksngram_freq$token1_freq/t) * (ksngram_freq$token2_freq/t)) * t
##WHEN RUNNING TRIGRAMS:##
ksngram_freq$expected_freq <- ((ksngram_freq$token1_freq/t) * (ksngram_freq$token2_freq/t) * (ksngram_freq$token3_freq/t)) * t
##WHEN RUNNING QUADRIGRAMS:##
ksngram_freq$expected_freq <- ((ksngram_freq$token1_freq/t) * (ksngram_freq$token2_freq/t) * (ksngram_freq$token3_freq/t) * (ksngram_freq$token4_freq/t)) * t

head(ksngram_freq)

######CALCULATING PMI FOR BIGRAMS######

#Defining PMI function
calculate_PMI <- function(observed_freq, token1_freq, token2_freq, expected_freq) {
  pmi <- log(observed_freq / expected_freq)
  return(pmi)
}

# Apply the function to create the PMI column
ksngram_freq$PMI <- with(ksngram_freq, calculate_PMI(observed_freq, token1_freq, token2_freq, expected_freq))

######PMI FOR TRIGRAMS######

#Defining PMI function for trigrams
calculate_PMI_trigram <- function(observed_freq, token1_freq, token2_freq, token3_freq, expected_freq) {
  pmi <- log(observed_freq / expected_freq)
  return(pmi)
}

ksngram_freq$PMI <- with(ksngram_freq, calculate_PMI_trigram(observed_freq, token1_freq, token2_freq, token3_freq, expected_freq))

######PMI FOR QUADRIGRAMS######

#Defining PMI function for quadrigrams
calculate_PMI_quadrigram <- function(observed_freq, token1_freq, token2_freq, token3_freq, token4_freq, expected_freq) {
  pmi <- log(observed_freq / expected_freq)
  return(pmi)
}

ksngram_freq$PMI <- with(ksngram_freq, calculate_PMI_quadrigram(observed_freq, token1_freq, token2_freq, token3_freq, token4_freq, expected_freq))

###############################

head(ksngram_freq)

#After running the above code for the entire data, rename:
ksngram_freq_all <- ksngram_freq

################################################################################
#Part 2: Making culture complex-specific subsets and generating PMI summary stats

#Subsetting the Maglemose data
maglemosedata <- subset(data, Chronology == "Maglemose")

#Subsetting the Kongemose data
kongemosedata <- subset(data, Chronology == "Kongemose")

#Subsetting the Ertebølle data
ertebølledata <- subset(data, Chronology == "Ertebølle")

#Generate skipgrams and PMI values for each of the subsets
#After running the above code (part 1) for the maglemose data, rename:
magle_ksngrams <- ksngram_freq
#After running the above code (part 1) for the kongemose data, rename:
konge_ksngrams <- ksngram_freq
#After running the above code (part 1) for the ertebølle data, rename:
erte_ksngrams <- ksngram_freq

head(ksngram_freq_all)
head(magle_ksngrams)
head(konge_ksngrams)
head(erte_ksngrams)

# Matching skipgrams occurring in each period with the overall stats
magle_ksngram_stats <- ksngram_freq_all[match(magle_ksngrams$ksngrams, ksngram_freq_all$ksngrams, nomatch = 0), ]
magle_ksngram_subset <- magle_ksngram_stats[magle_ksngram_stats$observed_freq > 2,]
nrow(magle_ksngram_subset)

konge_ksngram_stats <- ksngram_freq_all[match(konge_ksngrams$ksngrams, ksngram_freq_all$ksngrams, nomatch = 0), ]
konge_ksngram_subset <- konge_ksngram_stats[konge_ksngram_stats$observed_freq > 2,]
nrow(konge_ksngram_subset)

erte_ksngram_stats <- ksngram_freq_all[match(erte_ksngrams$ksngrams, ksngram_freq_all$ksngrams, nomatch = 0), ]
erte_ksngram_subset <- erte_ksngram_stats[erte_ksngram_stats$observed_freq > 2,]
nrow(erte_ksngram_subset)

#Sorting the three new culture complex-specific data sets by PMI in descending order
erte_ksngram_subset_sorted <- erte_ksngram_subset[order(erte_ksngram_subset$PMI), ]
konge_ksngram_subset_sorted <- konge_ksngram_subset[order(konge_ksngram_subset$PMI), ]
magle_ksngram_subset_sorted <- magle_ksngram_subset[order(magle_ksngram_subset$PMI), ]


#Subset to co-occuring motifs at 3 or more
ksngram_freq_subset <- subset(ksngram_freq, observed_freq > 2)

head(magle_ksngram_subset_sorted)
head(konge_ksngram_subset_sorted)
head(erte_ksngram_subset_sorted)
head(ksngram_freq_subset)

#Outputting the summary PMI stats
summary(ksngram_freq_subset$PMI)
summary(magle_ksngram_subset_sorted$PMI)
summary(konge_ksngram_subset_sorted$PMI)
summary(erte_ksngram_subset_sorted$PMI)


################################################################################
#Part 3: Making frequency- and PMI-sorted stacked bar plots#####################

#Before running the below code, make sure that the current 'ksngram_freq' is of
#the data set in its entirety (i.e., not subset to any culture complex)

#Subsetting by raw frequency (>9 is just an arbitrary threshold)
ksngram_freq_subset <- subset(ksngram_freq, observed_freq > 9)

#Sorting by observed freq.
ksngram_freq_subset <- ksngram_freq_subset %>% arrange(desc(observed_freq))


#Creating a new object, sorted by PMI in descending order
ksngram_PMI <- ksngram_freq %>% arrange(desc(PMI))

#Subsetting this object to exclude observed freq's below 3 (as these PMI's are
#the most dubious)
ksngram_PMI <- subset(ksngram_PMI, observed_freq > 2)

#Only including the highest 21 PMI scores (once again, an arbitrary number, simply
#set to 21 because there are only 21 bigrams with observed freq's above 9. Making
#this equal results in visually matching bar plots)
ksngram_PMI21 <- head(ksngram_PMI, 21)

head(ksngram_freq_subset)
head(ksngram_PMI21)


#To make the bar plots stacked, we need to add the chronology variable. As most
#of the skipgrams are attributed to more than one chronology level, we need to
#change the shape of the data, connecting each co-occurrence to its corresponding
#chronology

names(data)

#Changing the 'ksngrams' column contents to uppercase, in order to match it to
#the original data
ksngram_freq_subset$ksngrams <- toupper(ksngram_freq_subset$ksngrams)
ksngram_PMI21$ksngrams <- toupper(ksngram_PMI21$ksngrams)

#Making a new object with only the chronology variable and the individual motifs
chrongrams <-data[,c(11,31:289)]

createResultTable <- function(data) {
  result_tables <- list()
  skipgrams <- unique(data$ksngrams)
  
  for (skipgram in skipgrams) {
    filter_condition <- paste(strsplit(skipgram, " ")[[1]], "== 1", collapse = " & ")
    filtered_data <- chrongrams %>% filter(eval(parse(text = filter_condition)))
    
    result_table <- table(filtered_data$Chronology)
    result_tables[[skipgram]] <- data.frame(Skipgram = skipgram, Chronology = names(result_table), Frequency = as.numeric(result_table))
  }
  
  final_result <- do.call(rbind, result_tables)
  return(final_result)
}

skipgramsFreq <- createResultTable(ksngram_freq_subset)
skipgramsPMI <- createResultTable(ksngram_PMI21)

head(skipgramsFreq)
head(skipgramsPMI)


#Adding PMI values
ksngram_freq$ksngrams <- toupper(ksngram_freq$ksngrams)

skipgramsFreq$PMI <- ksngram_freq$PMI[match(skipgramsFreq$Skipgram, ksngram_freq$ksngrams)]
skipgramsPMI$PMI <- ksngram_freq$PMI[match(skipgramsPMI$Skipgram, ksngram_freq$ksngrams)]

head(skipgramsFreq)
head(skipgramsPMI)

#Extracting unique skipgrams from ksngram_freq_subset and set the order
skipgram_order <- unique(ksngram_freq_subset$ksngrams)

#Making sure to plot only one PMI value per stacked bar
summary_freq <- skipgramsFreq %>%
  distinct(Skipgram, .keep_all = TRUE)

#Plotting the skipgrams by frequency, stacked by chronology
skipgramsBYfrequency <- ggplot(skipgramsFreq, aes(x = Frequency, fill = Chronology, y = reorder(Skipgram, desc(match(Skipgram, skipgram_order))))) +
  geom_col(position = "stack") +
  geom_text(data = summary_freq, aes(label = sprintf("%.4f", PMI)), #Paste sprintf("%.4f", PMI instead on '""' to show the PMI)
            position = position_stack(vjust = 0), color = "black", size = 3) +
  scale_fill_manual(values = c("#FAD02E", "#D98888", "#E8BB8B", "#B4A1C1")) +
  xlab("Frequency") +
  ylab("Co-occurrence") +
  theme(plot.title = element_text(size = 13, face = "bold", hjust = 0.5),
        legend.text = element_text(size = 10),
        legend.title = element_blank(),
        legend.position = c(0.85, 0.15),
        legend.background = element_rect(fill = "transparent"),
        axis.text.x = element_text(vjust = 0.4, hjust = 1),
        text = element_text(size = 14, family = "Gill Sans MT")) +
  ggtitle("Most frequent _grams") #alter title depending on analysis level

skipgramsBYfrequency

#Doing the same thing, but plotting by PMI
# Extract unique Bigrams from skipgramsPMI and set the order
skipgram_order <- unique(skipgramsPMI$Skipgram)

# Making sure to plot only one PMI value per stacked bar
summary_pmi <- skipgramsPMI %>%
  distinct(Skipgram, .keep_all = TRUE)

skipgramsBYpmi <- ggplot(skipgramsPMI, aes(x = Frequency, fill = Chronology, y = reorder(Skipgram, PMI))) +
  geom_col(position = "stack") +
  geom_text(data = summary_pmi, aes(label = sprintf("%.4f", PMI)), #Paste sprintf("%.4f", PMI instead on '""' to show the PMI)
            position = position_stack(vjust = 0), color = "black", size = 3) +
  scale_fill_manual(values = c("#FAD02E", "#D98888", "#E8BB8B", "#B4A1C1")) +
  xlab("Frequency") +
  ylab("Co-occurrence") +
  theme(plot.title = element_text(size = 13, face = "bold", hjust = 0.5),
        axis.text.x = element_text(vjust = 1, hjust = 0.5),
        text = element_text(size = 14, family = "Gill Sans MT")) +
  ggtitle("_grams by highest PMI value")+
  guides(fill = FALSE)  #alter title depending on analysis level

skipgramsBYpmi

#The above plots can be repeated at different level of n-gram analysis


################################################################################
#Part 4: Making a heatmap of bigrams by their PMI###############################

#Ensure that 'ksngram_freq' is the entire data set
head(ksngram_freq)

#Subsetting to observed freqs of 3 or above (once again, due to the PMI values
#being increasingly uncertain at low observed freqs)
ksngram_freq <- subset(ksngram_freq, observed_freq > 2)


#Getting unique token values from ksngram_freq
tokens <- unique(c(as.character(ksngram_freq$token1), as.character(ksngram_freq$token2)))

#Creating empty matrix with token names as column and row names
m <- matrix(0, nrow = length(tokens), ncol = length(tokens), dimnames = list(tokens, tokens))

#Filling the matrix with PMI values from ksngram_freq
for (i in 1:nrow(ksngram_freq)) {
  row_index <- match(ksngram_freq[i, "token1"], tokens)
  col_index <- match(ksngram_freq[i, "token2"], tokens)
  m[row_index, col_index] <- ksngram_freq[i, "PMI"]
  m[col_index, row_index] <- ksngram_freq[i, "PMI"]
}

colnames(m)

# Defining the matrix and reorder the rows and columns
#(this manual chronological ordering is based on analyses not presented here)
m <- m[c("a14","b3","a24","b10","b2","b12","e4","a5","d5","b13","b4","b1","a1","a3","ant",
         "f12","e1","b6","e5","f5","f13","f38","d29","b8","d33","h1","g2","c1","f14","i1",
         "f35","zoo","d19","g1","d3","d15","f15","c12","c14","c5","c13","c4","c6","d7",
         "f55","g7","i13","i5","relief"), 
       c("a14","b3","a24","b10","b2","b12","e4","a5","d5","b13","b4","b1","a1","a3","ant",
         "f12","e1","b6","e5","f5","f13","f38","d29","b8","d33","h1","g2","c1","f14","i1",
         "f35","zoo","d19","g1","d3","d15","f15","c12","c14","c5","c13","c4","c6","d7",
         "f55","g7","i13","i5","relief")]

#Converting the matrix to a data frame for ggplot2
df <- as.data.frame.table(m)
names(df) <- c("token1", "token2", "PMI")

df

#Plotting the heatmap
heatmap <- ggplot(df, aes(x = token2, y = token1, fill = PMI)) +
  geom_tile(color = "gray90") +
  scale_fill_gradient2(low = "white", mid = "white", high = "red",
                       midpoint = 0.5, name = "PMI") +
  labs(x = "Token 2", y = "Token 1", title = "Co-occurrence matrix") +
  theme(axis.text.x = element_text(angle = 90, hjust = 1, vjust = 0.4),
        plot.title = element_text(size = 14, face = "bold", hjust = 0.5),
        text = element_text(family = "Gill Sans MT", size = 12),
        legend.text = element_text(size = 10)) +
  scale_x_discrete(limits = c("a14","b3","a24","b10","b2","b12","e4","a5","d5","b13","b4","b1","a1","a3","ant",
                              "f12","e1","b6","e5","f5","f13","f38","d29","b8","d33","h1","g2","c1","f14","i1",
                              "f35","zoo","d19","g1","d3","d15","f15","c12","c14","c5","c13","c4","c6","d7",
                              "f55","g7","i13","i5","relief"),
                   labels = toupper,
                   name = NULL,
                   drop = FALSE) +
  scale_y_discrete(limits = c("a14","b3","a24","b10","b2","b12","e4","a5","d5","b13","b4","b1","a1","a3","ant",
                              "f12","e1","b6","e5","f5","f13","f38","d29","b8","d33","h1","g2","c1","f14","i1",
                              "f35","zoo","d19","g1","d3","d15","f15","c12","c14","c5","c13","c4","c6","d7",
                              "f55","g7","i13","i5","relief"),
                   labels = toupper,
                   name = NULL,
                   drop = FALSE) +
  ggtitle("Motif co-occurrence heatmap by PMI")

heatmap

ggsave("heatmapPMI.png", plot = heatmap, width = 8, height = 8, units = "in", dpi = 300)

#The rest of the figure (adjusting axis labels and adding coloured boxes etc.)
#was then subsequently performed in the free image editing software 'GIMP'


################################################################################
#Part 5: Exporting the data tables##############################################

#Run the code down to line 144 at different levels of n-gram analysis, then
#export the data tables one at a time using the below code. We once again
#subset to observed freq's >2 and sort by PMI

ksngram_freq$token1 <- toupper(ksngram_freq$token1)
ksngram_freq$token2 <- toupper(ksngram_freq$token2)
##NEXT LINE IF WE'RE RUNNING TRIGRAMS##
ksngram_freq$token3 <- toupper(ksngram_freq$token3)
##NEXT LINE IF WE'RE RUNNING QUADRIGRAMS##
ksngram_freq$token4 <- toupper(ksngram_freq$token4)

head(ksngram_freq)

ksngram_freq <- ksngram_freq %>%
  select(token1, token2, token1_freq, token2_freq, observed_freq, expected_freq, PMI)
##NEXT LINE IF WE'RE RUNNING TRIGRAMS##
ksngram_freq <- ksngram_freq %>%
  select(token1, token2, token3, token1_freq, token2_freq, token3_freq, observed_freq, expected_freq, PMI)
##NEXT LINE IF WE'RE RUNNING QUADRIGRAMS##
ksngram_freq <- ksngram_freq %>%
  select(token1, token2, token3, token4, token1_freq, token2_freq, token3_freq, token4_freq, observed_freq, expected_freq, PMI)

head(ksngram_freq)

#Converting the data frame to a data table
ksngram_dt <- as.data.table(ksngram_freq)

#Subsetting to observed frequencies >2 and arranging by PMI
ksngram_dt <- subset(ksngram_dt, observed_freq > 2)
ksngram_dt <- ksngram_dt %>% arrange(desc(PMI))

#Writing the data table to a file (adjust name correspondingly)
fwrite(ksngram_dt, "_grams.csv")

#The aesthetic adjustments to the data tables were subsequently performed in Excel
